#include <XSUtil/Numerics/LinearInterp.h>
#include <XSUtil/Error/Error.h>

namespace Numerics {

  namespace Rebin {


    bool findFirstBins(const RealArray& currentBins, const RealArray& targetBins,
		       const Real FUZZY, size_t& currentStart, size_t& targetStart)
    {
      // Find first bin in targetBins which which is contained
      // in range of currentBins, and the bin in currentBins to
      // which it corresponds.  To conform with inibin.f function,
      // a target bin only partially contained on LOW end of 
      // currentBins is NOT accepted.  It IS accepted if it is
      // partially overlapping the HIGH end. If no target bin
      // is found, function returns false.
    
      bool isOverlap = false;
      currentStart = targetStart = 0;

      if (currentBins.size() && targetBins.size()) {

	size_t nCurrentBins = currentBins.size() - 1;
	size_t nTargetBins = targetBins.size() - 1;
	while (targetStart < nTargetBins && targetBins[targetStart]*(1.+FUZZY)
	       < currentBins[0]) ++targetStart;

	if (targetStart < nTargetBins) {
	  while (currentStart < nCurrentBins && currentBins[currentStart+1] 
		 < targetBins[targetStart]*(1.+FUZZY)) ++currentStart;
	  if (currentStart < nCurrentBins) isOverlap = true;
	}

      }

      return isOverlap;
    }



    void initializeBins(const RealArray& currentBins, const RealArray& targetBins,
			const Real FUZZY, size_t& currentStart, size_t& targetStart,
			IntegerVector& startBin, IntegerVector& endBin, 
			RealArray& startWeight, RealArray& endWeight)
    {
      // The assumption for original starting values is that 
      // currentStart and targetStart are the lowest values which
      // satisfy:
      //    targetBins[targetStart]*(1.0+FUZZY) >= currentBins[currentStart]
      // Therefore if there's any overlap between the two arrays,
      // at least one of currentStart,targetStart must be 0.
      size_t Nout (targetBins.size() - 1);
      size_t Nin  (currentBins.size() - 1);
      startWeight.resize(Nout,0);
      endWeight.resize(Nout,0);
      startBin.clear();
      endBin.clear();
      startBin.resize(Nout,0);
      endBin.resize(Nout,0);

      while ( targetStart < Nout ) {

	if ( targetBins[targetStart+1] <= currentBins[currentStart+1] ) {

	  endBin[targetStart] = 0;
	  startBin[targetStart] = currentStart;
	  startWeight[targetStart] 
	    = (targetBins[targetStart+1]- targetBins[targetStart])
	    / ( currentBins[currentStart+1] - currentBins[currentStart]);
	  if (targetBins[targetStart+1]*(1.0+FUZZY) >= currentBins[currentStart+1]) {
	    ++currentStart;
	  }
	  ++targetStart;

	} else {

	  startBin[targetStart] = currentStart;
	  startWeight[targetStart] 
	    = ( currentBins[currentStart+1] - targetBins[targetStart] )
	    / (currentBins[currentStart+1] - currentBins[currentStart]);
	  while ( currentStart < Nin 
		  && targetBins[targetStart+1] > currentBins[currentStart+1] ) {
	    ++currentStart;
	  }

	  if ( currentStart >= Nin ) {
	    endBin[targetStart] = Nin - 1;
	    endBin[targetStart] != startBin[targetStart] 
	      ?  endWeight[targetStart] = 1. 
	      : endBin[targetStart] = 0;
	    ++targetStart;       

	  } else {

	    endBin[targetStart] = currentStart;
	    endWeight[targetStart] 
	      = (targetBins[targetStart+1] - currentBins[currentStart])
	      / ( currentBins[currentStart+1] - currentBins[currentStart]);
	    if (targetBins[targetStart+1]*(1.0+FUZZY) >= currentBins[currentStart+1]) {
	      ++currentStart;
	    }
	    ++targetStart;
	  }

	}

	if ( currentStart >= Nin ) targetStart = Nout + 1;
      }
    }

    
    void rebin(const RealArray& inputArray, const IntegerVector& startBin, 
	       const IntegerVector& endBin, const RealArray& startWeight, 
	       const RealArray& endWeight, RealArray& outputArray, 
	       const Real lowValue, const Real highValue)
    {
      size_t Nout(outputArray.size());
      bool isLow(true);
      for (size_t j = 0; j < Nout; ++j) {

	if ( startBin[j] >= 0 ) {
	  isLow = false;
	  outputArray[j] = startWeight[j]*inputArray[startBin[j]];
	  if ( endBin[j] > startBin[j] ) {
	    for ( int k = startBin[j] + 1; k < endBin[j]; ++k) {
	      outputArray[j] += inputArray[k];
	    } 
	    outputArray[j] += endWeight[j]*inputArray[endBin[j]];
	  }      

	} else {

	  if ( isLow ) {
	    outputArray[j] = lowValue;
	  } else {
	    outputArray[j] = highValue;
	  }

	}

      }

    }

   void interpolate(const RealArray& inputArray,const IntegerVector& startBin, 
		    const IntegerVector& endBin, const RealArray& startWeight, 
		    const RealArray& endWeight, RealArray& outputArray, 
		    bool exponential, const Real lowValue, const Real highValue)
   {
      size_t Nout (outputArray.size());
      bool isLow = true;

      for ( size_t j = 0; j < Nout; ++j) {

	Real denominator (0.);
	outputArray[j] = startWeight[j]*inputArray[startBin[j]];
	denominator += startWeight[j];
	if ( endBin[j] > startBin[j] ) {
	  for ( int k = startBin[j] + 1; k < endBin[j]; ++k) {
	    outputArray[j] += inputArray[k];
	    denominator += 1;
	  }
	  outputArray[j] += endWeight[j]*inputArray[endBin[j]];
	  denominator += endWeight[j];
	}

	if ( denominator > 0 ) {
	  isLow = false;
	  outputArray[j] /= denominator;
	} else {
	  if ( isLow ) {
	    outputArray[j] = ( exponential ? -log(lowValue) :  lowValue) ;
	  } else {
	    outputArray[j] = ( exponential ? -log(highValue) :  highValue) ;
	  }
	}

      }

      if ( exponential ) {
	for (size_t j = 0; j < Nout; ++j) { 
	  outputArray[j] > 100. 
	    ?  outputArray[j] = 0.
	    : outputArray[j] = exp(-outputArray[j]);
	}           
      }

   }

    void gainRebin(const RealArray& inputArray, const IntegerVector &startBin,
		   const IntegerVector& endBin, const RealArray& startWeight,
		   const RealArray& endWeight, RealArray& outputArray)
    {
      // ASSUMPTION: All arrays passed here are the SAME SIZE
      size_t sz = outputArray.size();

      // Partial check.
      if (sz != inputArray.size() || sz != startBin.size()) {
	throw RedAlert("Array size mismatch occured in gainRebin function.");
      }

      for (size_t i=0; i<sz; ++i) {
	Real total = .0;
	Real& output_i = outputArray[i];
	const int startBin_i = startBin[i];
	const int endBin_i = endBin[i];
	output_i = startWeight[i]*inputArray[startBin_i];
	total += startWeight[i];
	if (endBin_i > startBin_i) {
	  for (int j=startBin_i+1; j<endBin_i; ++j) {
	    // If we're here, output bin spreads over more than
	    // just 2 input bins.
	    output_i += inputArray[j];
	    total += 1.0;
	  }
	  output_i += endWeight[i] *inputArray[endBin_i];
	  total += endWeight[i];
	}
	if (total > .0) {
	  output_i /= total;
	} else {
	  output_i = 1.0;
	}
      }
    }

    void linInterpInteg(const RealArray& currentPoints, const RealArray& inputdValues, 
			const RealArray& targetBins, RealArray& outputValues, 
			const Real lowValue, const Real highValue)
    {
      // This calculates the integrated outputValues[i] between targetBins[i] and
      // targetBins[i+1] by linearly interpolating on the input inputdValues which
      // has values at currentPoints.

      outputValues.resize(targetBins.size()-1);
      outputValues = 0.0;

      
      // Check for cases no overlap

      if ( currentPoints[currentPoints.size()-1] < targetBins[0] ) {
	outputValues = highValue;
	return;
      }
      if ( currentPoints[0] > targetBins[targetBins.size()-1] ) {
	outputValues = lowValue;
	return;
      }
      
      // Find the first element in currentPoints which lies within the
      // targetBins and the start element of the bin in which it lies.
      
      size_t currentPos(0), targetPos(0);
      const size_t nCurrent(currentPoints.size());
      const size_t nTarget(targetBins.size());

      while ( currentPoints[currentPos] < targetBins[targetPos] &&
	      currentPos < nCurrent-2 ) {
	currentPos++;
      }
      while ( currentPoints[currentPos] > targetBins[targetPos] &&
	      targetPos < nTarget-2 ){
	targetPos++;
      }
      if ( targetPos > 0 ) targetPos--;

      // currentPoints[currentPos] should now lie between targetBins[targetPos]
      // and targetBins[targetPos+1] unless currentPoints[currentPos] is below
      // targetBins[0] or above targetBins[nTarget-2].

      Real point = currentPoints[currentPos];
      Real pointNext = currentPoints[currentPos+1];
      Real dVal = inputdValues[currentPos];
      Real dValNext = inputdValues[currentPos+1];

      Real targL = targetBins[targetPos];
      Real targH = targetBins[targetPos+1];

      // if targetPos > 0 we need to start by calculating outputValues
      // for earlier bins

      if ( targetPos > 0 ) {

	if ( currentPos == 0 ) {

	  // if currentPos=0 set all the lower output bins to lowValue

	  for (size_t i=0; i<targetPos; i++) {
	    outputValues[i] = lowValue*(targetBins[i+1]-targetBins[i]);
	  }

	} else {

	  // all the bins below targetPos lie between currentPos and
	  // currentPos-1. we know by construction that pointLast is
	  // below targetBins[0].
	  Real pointLast = currentPoints[currentPos-1];
	  Real delPoint = point - pointLast;
	  Real dValLast = inputdValues[currentPos-1];
	  for (size_t i=0; i<targetPos; i++) {
	    Real tLow = targetBins[i];
	    Real tHigh = targetBins[i+1];
	    Real tMid = 0.5*(tLow+tHigh);
	    outputValues[i] += (tHigh-tLow)
	      * ( dValLast*(point-tMid) + dVal*(tMid-pointLast) ) / delPoint;

	  }

	}

      }

      // if currentPos > 0 then we need to include the contribution from
      // targL to point for the targetPos bin. by construction we know
      // that currentPoints[currentPos-1] is less than targL

      if ( currentPos > 0 ) {
	Real pointLast = currentPoints[currentPos-1];
	Real dValLast = inputdValues[currentPos-1];
	outputValues[targetPos] += 0.5 * (point-targL)
	  * ( dValLast*(point-targL) + dVal*(point-pointLast+targL-pointLast) ) / (point-pointLast);
      }

      // loop till we run out of targetBins bins or currentPoints points

      while ( targetPos < nTarget-1 && currentPos < nCurrent-1 ) {

	// if the next input point lies in the next output bin

	if ( pointNext >= targH ) {

	  if ( point <= targL ) {

	    Real targM = 0.5*(targL+targH);
	    outputValues[targetPos] += (targH-targL)
	      * ( dVal*(pointNext-targM) + dValNext*(targM-point) ) / (pointNext-point);

	  } else {

	    outputValues[targetPos] += 0.5 * (targH-point)
	      * ( dVal*(pointNext-point+pointNext-targH) + dValNext*(targH-point) ) / (pointNext-point);

	  }

	  targetPos++;
          if (targetPos < nTarget-1) {
	     targL = targH;
	     targH = targetBins[targetPos+1];
          }

	  // now do the case of the next input point lies in the current output bin

	} else if ( pointNext < targH ) {

	  if ( point <= targL ) {

	    outputValues[targetPos] += 0.5 * (pointNext-targL)
	      * ( dVal*(pointNext-targL) + dValNext*(pointNext-point+targL-point) ) / (pointNext-point);

	  } else {

	    outputValues[targetPos] += 0.5 * (pointNext-point) * (dVal + dValNext);

	  }

	  currentPos++;
          if (currentPos < nCurrent-1)
          {
	     point = pointNext;
	     pointNext = currentPoints[currentPos+1];
	     dVal = dValNext;
	     dValNext = inputdValues[currentPos+1];
          }
	}

      }

      // Check whether we have got to the end of the currentPoints array before we
      // have done all the output bins

      if ( currentPos == nCurrent-1 && targetPos < nTarget-1 ) {
	outputValues[targetPos] += dVal*(targH-point);
      }

      // Set any remaining output bins to highValue

      for (size_t i=targetPos+1; i<outputValues.size(); i++) {
	outputValues[i] = highValue*(targetBins[i+1]-targetBins[i]);
      }

      return;
    }

  }  // namespace Rebin

}  // namespace Numerics
